#include "BuddhaModel.h"
#include "BuddhaXNAPrerequisites.h"
#include "BuddhaGraphicsDevice.h"
#include "BuddhaTexture2D.h"
#include "BuddhaBasicEffect.h"
#include "BuddhaVertexDeclaration.h"
#include "BuddhaVertexBuffer.h"
#include "BuddhaIndexBuffer.h"
#include "BuddhaModelMesh.h"
#include "BuddhaModelMeshPart.h"
#include "BuddhaStringConvert.h"
#include "BuddhaUtils.h"
#include "BuddhaHardwareEffect.h"
#include "BuddhaModelBone.h"
#include "BuddhaXNAException.h"

namespace BuddhaXNA
{
	Model::Model(GraphicsDevice* device)
		: m_pDevice(device)
	{
	}
	//------------------------------------------------------------------------
	Model::~Model()
	{
		for (ModelMeshList::iterator itr = mModelMeshList.begin(), end = mModelMeshList.end();
			itr != end; ++itr)
		{
			SAFE_DELETE(*itr);
		}
	}
	//------------------------------------------------------------------------
	int Model::GetNumModelMeshes(void) const
	{
		return static_cast<int>(mModelMeshList.size());
	}
	//------------------------------------------------------------------------
	ModelMesh* Model::GetModelMesh(int idx) const
	{
		if (idx >= 0 && idx < GetNumModelMeshes())
			return mModelMeshList[idx];

		throw ArgumentOutOfRangeException();
	}
	//------------------------------------------------------------------------
	ModelPtr Model::_fromFile(GraphicsDevice* device, LPCWSTR filePath)
	{
		HRESULT hr;
		LPD3DXBUFFER p_buf;

		/// DirectX Mesh Data for constructing a Buddha Model.
		LPD3DXMESH t_oldMesh, t_newMesh;

		if (device == NULL)
			throw "The graphic device is NULL";

		//
		// t@CT
		//
		String truePath = Utils::FindMediaFile(String(filePath));
		if (truePath == L"")
			throw "The file is not found.";

		Model* model = new Model(device);

		DWORD numMaterials = 0;

		hr = D3DXLoadMeshFromX(
			truePath.c_str(),
			D3DXMESH_SYSTEMMEM,
			device->_getD3D9Device(),
			NULL,
			&p_buf,
			NULL,
			&numMaterials,
			&t_oldMesh
			);

		if (FAILED(hr)) {
			delete model;	// TODO O
			return ModelPtr(NULL);
		}

		D3DVERTEXELEMENT9 pDecl[MAX_FVF_DECL_SIZE];
		t_oldMesh->GetDeclaration(pDecl);

		//
		// bVf[^œKisvȂj
		//
		hr = t_oldMesh->CloneMesh(D3DXMESH_MANAGED, pDecl, device->_getD3D9Device(), &t_newMesh);
		SAFE_RELEASE(t_oldMesh);

		const DWORD fvf = t_newMesh->GetFVF();

		// _錾
		VertexDeclarationPtr declaration(new VertexDeclaration(device, pDecl));

		//D3DXComputeNormals(t_newMesh, NULL);

		//--------------------------------------------------
		// _obt@̎擾 & Buddha Œ_obt@̍쐬
		//--------------------------------------------------
		LPVOID pBuffer;
		t_newMesh->LockVertexBuffer(D3DLOCK_READONLY, &pBuffer);
		t_newMesh->UnlockVertexBuffer();

		VertexBuffer* pVB = new VertexBuffer(
			device,
			t_newMesh->GetNumVertices() * t_newMesh->GetNumBytesPerVertex(),
			ResourceUsage::None,
			fvf,
			ResourceManagementMode::Automatic);

		pVB->SetData(pBuffer, t_newMesh->GetNumVertices() * t_newMesh->GetNumBytesPerVertex());

		VertexBufferPtr vertexBuffer(pVB);
		
		//------------------------------------------------------------------
		// CfbNXobt@̎擾 & Buddha ŃCfbNXobt@̍쐬
		//------------------------------------------------------------------
		t_newMesh->LockIndexBuffer(D3DLOCK_READONLY, &pBuffer);
		t_newMesh->UnlockIndexBuffer();

		IndexBuffer* pIB = new IndexBuffer(
			device,
			t_newMesh->GetNumFaces() * sizeof(SHORT) * 3,	// TODO this is for only short.
			ResourceUsage::None,
			ResourceManagementMode::Automatic,
			IndexElementSize::SixteenBits);

		pIB->SetData((SHORT*)pBuffer, t_newMesh->GetNumFaces() * 3);

		IndexBufferPtr indexBuffer(pIB);

		//------------------
		// bV\zJn
		//------------------

		// ̃[hV[NGXł́AqGL[Ȃ̂łP
		model->mModelMeshList.resize(1);
		ModelMesh* mesh = new ModelMesh(model, 0);
		model->mModelMeshList[0] = mesh;
		model->mModelMeshList[0]->_fvf = fvf;

		//------------------------
		// }eA̎擾Ɗi[
		//------------------------
		D3DXMATERIAL* pMaterials = (D3DXMATERIAL*)p_buf->GetBufferPointer();
		mesh->mMaterialList.resize(numMaterials);
		for (DWORD i = 0; i < numMaterials; i++)
		{
			mesh->mMaterialList[i] = new ModelMeshMaterial();

			mesh->mMaterialList[i]->_D3D9Material = pMaterials[i].MatD3D;
			//memcpy_s(mesh->mMaterialList[i]->_D3D9Material, sizeof(D3DMATERIAL9), pMaterials[i].MatD3D, sizeof(D3DMATERIAL9));

			if (pMaterials[i].pTextureFilename != NULL)
			{
				// eNX`̃[h
				String textureFilename = AsciiStringConvert::ToString(pMaterials[i].pTextureFilename);
				mesh->mMaterialList[i]->TexturePtr = Texture2D::FromFile(device, textureFilename.c_str());
			}
		}

		// _ & CfbNXobt@ZbgiȌ͂ɔCj
		mesh->mVertexBuffer = vertexBuffer;
		mesh->mNumVertexes = t_newMesh->GetNumVertices();
		mesh->mIndexBuffer = indexBuffer;

		// Agr[Ƃ𒲂ׂ
		DWORD t_AttributeTableSize;
		t_newMesh->GetAttributeTable(NULL, &t_AttributeTableSize);
		
		D3DXATTRIBUTERANGE* pAttributes = new D3DXATTRIBUTERANGE[t_AttributeTableSize];

		t_newMesh->GetAttributeTable(pAttributes , &t_AttributeTableSize);

		// `敪KvȐbVp[g쐬
		// TODO check mMeshPartsCount == t_AttributeTableSize
		mesh->mModelMeshPartList.resize(numMaterials);

		//-------------------------
		// ebVp[gZbg
		//-------------------------
		for (DWORD i = 0; i < numMaterials; i++)
		{
			D3DXATTRIBUTERANGE t_range = pAttributes[i];

			ModelMeshPart* pPart = new ModelMeshPart();
			mesh->mModelMeshPartList[i] = pPart;

			pPart->mAttributeId = pAttributes[i].AttribId;
			pPart->mStartIndex = pAttributes[i].FaceStart * 3;
			pPart->mPrimitiveCount = pAttributes[i].FaceCount;
			pPart->mBaseVertex = pAttributes[i].VertexStart;
			pPart->mNumVertices = pAttributes[i].VertexCount;

			pPart->mVertexDeclaration = declaration;
			pPart->mVertexStride = t_newMesh->GetNumBytesPerVertex();
		}

		delete[] pAttributes;

		SAFE_RELEASE(t_newMesh);

		return ModelPtr(model);
	}
	//------------------------------------------------------------------------
	ModelPtr Model::FromFile(GraphicsDevice* device, LPCWSTR filePath)
	{
		ModelPtr model = _fromFile(device, filePath);
		if (model.IsNull())
			return model;

		for (int i = 0; i < model->GetNumModelMeshes(); i++)
		{
			EffectPtr effect(new HardwareEffect(device));
			model->GetModelMesh(i)->mEffectList.push_back(effect);
		}

		return model;
	}
	//------------------------------------------------------------------------
	ModelPtr Model::FromFile(GraphicsDevice* device, LPCWSTR filePath, EffectPtr effect)
	{
		ModelPtr model = FromFile(device, filePath);
		if (model.IsNull())
			return model;

		for (int i = 0; i < model->GetNumModelMeshes(); i++)
		{
			model->GetModelMesh(i)->mEffectList.push_back(effect);
		}

		return model;
	}
	//------------------------------------------------------------------------
	unsigned int Model::GetNumModelBones(void) const
	{
		return static_cast<unsigned int>(mModelBoneList.size());
	}
	//------------------------------------------------------------------------
	ModelBone* Model::GetModelBone(unsigned int idx) const
	{
		if (idx >= 0 && idx < mModelBoneList.size())
			return mModelBoneList[idx];

		throw ArgumentOutOfRangeException();
	}
}
